//
//  Base64 En/Decoder
//
//  Written in 2003 by Yves Goergen
//  Internet: software.unclassified.de
//
//  Usage: base64 -d|-e input [output]
//
//  -d       decode file
//  -e       encode file
//  input    file to convert*
//  output   file to write into*
//           if this is left out, default value is:
//           -d: input - ".b64" (if present)
//               input + ".bin" (else)
//           -e: input + ".b64"
//
//  *) files can be replaced by '=' for access to clipboard.
//
//  Returns 0 on success, something else on error.
//
//  This tiny programme may be a little memory wasty but
//  it worked fine on a 90 meg testfile in less than a
//  minute on a rather slow machine and decoded even faster.
//
//  Base64 En/Decoder is open-source, freeware and may so
//  be copied and/or changed freely.
//  If you make any changes or anhancements or find a bug
//  in here, please let me know.
//


//  include some header files
//
#include <windows.h>
#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <io.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <ctype.h>


//  function prototypes
//
int POS(char c);
char *base64encode(const void *buf, int size);
int base64decode(char *s, void *data);


//  entry point
//
int main(int argc, char *argv[], char *envp[])
{
	// parse cmdline
	char *infilename, outfilename[260];
	bool encode = false;
	if (argc < 3 || argc > 4)   // expect 2 to 3 arguments
	{
		printf("Base64 En/Decoder (2003) software.unclassified.de\r\n\r\n");
		printf("base64 -d|-e input [output]\r\n");
		printf("'=' for clipboard access\r\n");
		return 1;
	}
	if (!stricmp(argv[1], "-d"))   // "decode"
	{
		infilename = argv[2];
		if (argc == 4)
		{
			strcpy(outfilename, argv[3]);   // output filename given
		}
		else
		{
			if (strcmp(argv[2], "="))
				strcpy(outfilename, argv[2]);
			else
				strcpy(outfilename, "_base64_tmp");
			if (!strnicmp(outfilename + strlen(outfilename) - 4, ".b64", 4))
			{
				outfilename[strlen(outfilename) - 4] = 0;   // cut ".b64" if present
			}
			else
			{
				strcat(outfilename, ".bin");   // else add ".bin"
			}
		}
	}
	else if (!stricmp(argv[1], "-e"))   // "encode"
	{
		encode = true;
		infilename = argv[2];
		if (argc == 4)
		{
			strcpy(outfilename, argv[3]);   // output filename given
		}
		else
		{
			if (strcmp(argv[2], "="))
				strcpy(outfilename, argv[2]);
			else
				strcpy(outfilename, "_base64_tmp");
			strcat(outfilename, ".b64");   // add ".b64"
		}
	}
	else
	{
		printf("Base64 En/Decoder (2003) software.unclassified.de\r\n\r\n");
		printf("base64 -d|-e input [output]\r\n");
		printf("'=' for clipboard access\r\n");
		return 2;
	}

	int infile, outfile;
	long insize;
	long outsize;
	long current, next, len, x;
	char *bindata, *b64data, *b64data_m;

	HGLOBAL hglb;
	LPTSTR lptstr;

	if (strcmp(infilename, "="))
	{
		// open input file
		infile = _open(infilename, _O_RDONLY | _O_BINARY);
		if (!infile)
		{
			printf("Error: can't open infile\r\n");
			return 3;
		}
		
		// get input filesize
		lseek(infile, 0, SEEK_END);
		insize = tell(infile);
		lseek(infile, 0, SEEK_SET);

		// read input file
		if (encode)
		{
			bindata = (char *) malloc(insize);
			len = _read(infile, bindata, insize);
			if (len != insize)
			{
				free(bindata);
				printf("Error: can't read infile\r\n");
				return 4;
			}
		}
		else  /* !encode */
		{
			b64data_m = (char *) malloc(insize + 1);  // multiline
			b64data = (char *) malloc(insize + 1);    // one-string
			len = _read(infile, b64data_m, insize);
			if (len != insize)
			{
				free(b64data);
				free(b64data_m);
				printf("Error: can't read infile\r\n");
				return 4;
			}
			b64data_m[insize] = 0;
		}
	}
	else  /* = */
	{
		if (!OpenClipboard(0))
		{
			printf("Error: can't open clipboard\r\n");
			return 8;
		}
		if (!IsClipboardFormatAvailable(CF_TEXT))
		{
			printf("Error: no text in the clipboard\r\n");
			return 9;
		}

		hglb = GetClipboardData(CF_TEXT);
		if (hglb)
		{
			lptstr = (char *) GlobalLock(hglb);
			if (lptstr)
			{
				insize = strlen(lptstr);

				if (encode)
				{
					bindata = (char *) malloc(insize);
					memcpy(bindata, lptstr, insize);
				}
				else  /* !encode */
				{
					b64data_m = (char *) malloc(insize + 1);  // multiline
					b64data = (char *) malloc(insize + 1);    // one-string
					memcpy(b64data_m, lptstr, insize);
					b64data_m[insize] = 0;
				}

				GlobalUnlock(hglb);
			}
		}
		CloseClipboard();
	}

	if (encode)
	{
		// encode data
		b64data = base64encode(bindata, insize);
		outsize = strlen(b64data);

		if (strcmp(outfilename, "="))
		{
			// open output file
			outfile = _open(outfilename, _O_RDWR | _O_CREAT | _O_TRUNC | _O_BINARY, _S_IWRITE | _S_IREAD);
			if (!outfile)
			{
				free(bindata);
				free(b64data);
				printf("Error: can't open outfile\r\n");
				return 5;
			}

			// write to outfile, with line-splitting at 76 chars
			len = 0;
			for (current = 0; current < outsize; current += 76)
			{
				x = (outsize - current > 76 ? 76 : outsize - current);
				next = _write(outfile, b64data + current, x);
				next += _write(outfile, "\r\n", 2);
				if (next != x + 2)
				{
					free(bindata);
					free(b64data);
					printf("Error: can't write to outfile\r\n");
					return 6;
				}
				
				len += 2;
			}
		}
		else  /* = */
		{
			if (!OpenClipboard(0))
			{
				printf("Error: can't open clipboard\r\n");
				return 8;
			}
			EmptyClipboard();

			hglb = GlobalAlloc(GMEM_MOVEABLE, outsize  + (outsize / 76) * 2      + 2       + 1);
			//                                datasize + #of_linebreaks * 2chars + add_lbr + null
			if (!hglb)
			{
				printf("Error: can't allocate global memory\r\n");
				return 10;
			}

			// prepare base64 data, with line-splitting at 76 chars
			lptstr = (LPTSTR) GlobalLock(hglb);
			len = 0;
			for (current = 0; current < outsize; current += 76)
			{
				x = (outsize - current > 76 ? 76 : outsize - current);   // number of bytes for current data block
				memcpy(lptstr + current + len, b64data + current, x);
				lptstr[current + len + x] = '\r';
				lptstr[current + len + x + 1] = '\n';
				
				len += 2;
			}

			lptstr[outsize + len] = 0;
			GlobalUnlock(hglb);

			SetClipboardData(CF_TEXT, hglb);

			CloseClipboard();
		}

		// display summary
		if (!strcmp(infilename, "=")) strcpy(infilename, "(clipboard)");
		if (!strcmp(outfilename, "=")) strcpy(outfilename, "(clipboard)");
		printf("%s (%lu bytes) base64-encoded into %s (%lu bytes)\r\n",
			infilename, insize, outfilename, outsize + len);
	}
	else  /* !encode */
	{
		// remove new-lines
		len = 0;
		for (current = 0; current < insize; current++)
		{
			next = strcspn(b64data_m + current, " \r\n\t");
			if (next)
			{
				memcpy(b64data + len, b64data_m + current, next);
				current += next;
				len += next;
			}
		}
		free(b64data_m);
		b64data[len] = 0;

		// decode data
		bindata = (char *) malloc(insize + 1);
		outsize = base64decode(b64data, bindata);
		if (outsize == -1)
		{
			free(bindata);
			free(b64data);
			printf("Error: infile is not valid Base64\r\n");
			return 7;
		}

		if (strcmp(outfilename, "="))
		{
			// open output file
			outfile = _open(outfilename, _O_RDWR | _O_CREAT | _O_TRUNC | _O_BINARY, _S_IWRITE | _S_IREAD);
			if (!outfile)
			{
				free(bindata);
				free(b64data);
				printf("Error: can't open outfile\r\n");
				return 5;
			}

			// write to output file
			next = _write(outfile, bindata, outsize);
			if (next != outsize)
			{
				free(bindata);
				free(b64data);
				printf("Error: can't write to outfile\r\n");
				return 6;
			}
		}
		else  /* = */
		{
			if (!OpenClipboard(0))
			{
				printf("Error: can't open clipboard\r\n");
				return 8;
			}
			EmptyClipboard();

			hglb = GlobalAlloc(GMEM_MOVEABLE, outsize + 1); 
			if (!hglb)
			{
				printf("Error: can't allocate global memory\r\n");
				return 10;
			}

			// copy bin data
			lptstr = (LPTSTR) GlobalLock(hglb);
			memcpy(lptstr, bindata, outsize);
			lptstr[outsize] = 0;   // only relevant for text-only data...
			GlobalUnlock(hglb);

			SetClipboardData(CF_TEXT, hglb);

			CloseClipboard();
		}

		// display summary
		if (!strcmp(infilename, "=")) strcpy(infilename, "(clipboard)");
		if (!strcmp(outfilename, "=")) strcpy(outfilename, "(clipboard)");
		printf("%s (%lu bytes) base64-decoded into %s (%lu bytes)\r\n",
			infilename, insize, outfilename, outsize);
	}

	// close files and free memory
	_close(infile);
	_close(outfile);

	free(bindata);
	free(b64data);

	return 0;
}


//  single base64 character conversion
//
int POS(char c)
{
	if (c == '=') return 64;
	if (isupper(c)) return c - 'A';
	if (islower(c)) return c - 'a' + 26;
	if (isdigit(c)) return c - '0' + 52;
	if (c == '+') return 62;
	if (c == '/') return 63;
	return -1;
}


//  base64 encoding
//
//  buf:     binary input data
//  size:    size of input (bytes)
//  return:  base64-encoded string (null-terminated)
//           memory for output will be allocated here, free it later
//
char *base64encode(const void *buf, int size)
{
	char base64[] = "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/";
	char *str = (char *) malloc((size + 3) * 4 / 3 + 1);
	char *p = str;
	unsigned char *q = (unsigned char *) buf;
	int i = 0;
	int c;

	while (i < size)
	{
		c = q[i++];
		c *= 256;
		if (i < size) c += q[i];
		i++;
		c *= 256;
		if (i < size) c += q[i];
		i++;
		p[0] = base64[(c & 0x00fc0000) >> 18];
		p[1] = base64[(c & 0x0003f000) >> 12];
		p[2] = base64[(c & 0x00000fc0) >> 6];
		p[3] = base64[(c & 0x0000003f) >> 0];
		if (i > size) p[3] = '=';
		if (i > size + 1) p[2] = '=';
		p += 4;
	}
	*p = 0;

	return str;
}


//  base64 decoding
//
//  s:       base64 string, must be null-terminated
//  data:    output buffer for decoded data
//  return:  size of decoded data or error (-1)
//
int base64decode(char *s, void *data)
{
	char *p;
	unsigned char *q;
	int n[4];

	if (strlen(s) % 4) return -1;
	q = (unsigned char *) data;
	for (p = s; *p; p += 4)
	{
		n[0] = POS(p[0]);
		n[1] = POS(p[1]);
		n[2] = POS(p[2]);
		n[3] = POS(p[3]);
		if ((n[0] | n[1] | n[2] | n[3]) < 0) return -1;
		if (n[0] == 64 || n[1] == 64) return -1;
		if (n[2] == 64 && n[3] < 64) return -1;

		q[0] = (n[0] << 2) + (n[1] >> 4);
		if (n[2] < 64) q[1] = ((n[1] & 15) << 4) + (n[2] >> 2);
		if (n[3] < 64) q[2] = ((n[2] & 3) << 6) + n[3];
		q += 3;
	}
	q -= (n[2] == 64) + (n[3] == 64);

	return q - (unsigned char *) data;
}
